let _ = Mltop.add_known_module "coq-quickchick.plugin"

# 3 "driver.mlg"
 
open Libnames
open Util
open Constrexpr
open Names
open Stdarg
open Error

type derivation = SimpleDer of SimplDriver.derivable list
                | DepDer of DepDriver.derivable

let simpl_dispatch ind classes name1 name2 =
  let ind_name = match ind with
    | { CAst.v = CRef (r, _); _ } -> string_of_qualid r
    | _ -> failwith "Implement me for functions"
  in
  List.iter (fun cn -> SimplDriver.derive cn ind (SimplDriver.mk_instance_name cn ind_name) name1 name2) classes

let class_assoc_opts = [ ("GenSized"                 , SimpleDer [SimplDriver.GenSized])
                       ; ("EnumSized"                , SimpleDer [SimplDriver.EnumSized])                       
                       ; ("Shrink"                   , SimpleDer [SimplDriver.Shrink])
                       ; ("Arbitrary"                , SimpleDer [SimplDriver.GenSized; SimplDriver.Shrink])
                       ; ("Show"                     , SimpleDer [SimplDriver.Show])
                       ; ("Sized"                    , SimpleDer [SimplDriver.Sized])
                       ; ("DecOpt"                   , DepDer DepDriver.DecOpt)
                       ; ("ArbitrarySizedSuchThat"   , DepDer DepDriver.GenSizedSuchThat)
                       ; ("GenSizedSuchThat"         , DepDer DepDriver.GenSizedSuchThat)                       
                       ; ("EnumSizedSuchThat"        , DepDer DepDriver.EnumSizedSuchThat)
                       ; ("Generator"                , DepDer DepDriver.GenSizedSuchThat)
                       ; ("Enumerator"               , DepDer DepDriver.EnumSizedSuchThat)
                       ; ("Checker"                  , DepDer DepDriver.DecOpt)
                       ]

let class_assoc_table =
  let h = Hashtbl.create (List.length class_assoc_opts) in
  List.iter (fun (kwd, tok) -> Hashtbl.add h kwd tok) class_assoc_opts;
  h

let dispatch cn ind name1 name2 =
  let convert_reference_to_string c =
    match c with
    | {CAst.v = CRef (r, _) ; _} -> string_of_qualid r
    | _ -> failwith "Usage: Derive <class_name> for <inductive_name> OR  Derive (<class_name>, ... , <class_name>) for <inductive_name>" in
  let ss = match cn.CAst.v with
     | CNotation (_, _, ([a],[b],_,_)) ->
         let c = convert_reference_to_string a in
         let cs = List.map convert_reference_to_string b in
         c :: cs
     | _ -> [convert_reference_to_string cn]
  in

  let get_class_names s =
    try Hashtbl.find class_assoc_table s
    with Not_found -> begin
        (* TODO: Figure out how to pretty print in a failwith... *)
        failwith ( "Not a valid class name. Expected one of : " ^ (String.concat " , " (List.map fst class_assoc_opts)))
      end
  in

  let class_names =
    match ss with
    | s::ss -> List.fold_left (fun der s ->
                               match der, get_class_names s with
                               | SimpleDer ds1, SimpleDer ds2 -> SimpleDer (ds1 @ ds2)
                               | DepDer ds1, DepDer ds2 ->
                                  qcfail "Implement dependent derive for multiple classes"
                              ) (get_class_names s) ss
    | _ -> qcfail "At least one class name expected" in

  match class_names with
  | SimpleDer classes -> simpl_dispatch ind classes name1 name2
  | DepDer class_name -> DepDriver.dep_dispatch ind class_name
;;                       

let merge ind1 ind2 ind3 =
  MergeTypes.merge ind1 ind2 ind3

let merge_test ind1  =
  MergeTypes.merge_test ind1


let () = Vernacextend.static_vernac_extend ~plugin:(Some "coq-quickchick.plugin") ~command:"Merge" ~classifier:(fun _ -> Vernacextend.classify_as_sideeff) ?entry:None 
         [(Vernacextend.TyML (false, Vernacextend.TyTerminal ("Merge", 
                                     Vernacextend.TyNonTerminal (Extend.TUentry (Genarg.get_arg_tag wit_constr), 
                                     Vernacextend.TyTerminal ("With", 
                                     Vernacextend.TyNonTerminal (Extend.TUentry (Genarg.get_arg_tag wit_constr), 
                                     Vernacextend.TyTerminal ("As", Vernacextend.TyNonTerminal (
                                                                    Extend.TUentry (Genarg.get_arg_tag wit_constr), 
                                                                    Vernacextend.TyNil)))))), 
         (let coqpp_body ind1 ind2 ind
         () = Vernacextend.vtdefault (fun () -> 
# 87 "driver.mlg"
      merge ind1 ind2 ind 
              ) in fun ind1
         ind2 ind ?loc ~atts () -> coqpp_body ind1 ind2 ind
         (Attributes.unsupported_attributes atts)), None))]

let () = Vernacextend.static_vernac_extend ~plugin:(Some "coq-quickchick.plugin") ~command:"MergeTest" ~classifier:(fun _ -> Vernacextend.classify_as_sideeff) ?entry:None 
         [(Vernacextend.TyML (false, Vernacextend.TyTerminal ("MergeTest", 
                                     Vernacextend.TyNonTerminal (Extend.TUentry (Genarg.get_arg_tag wit_constr), 
                                     Vernacextend.TyNil)), (let coqpp_body ind1
                                                           () = Vernacextend.vtdefault (fun () -> 
                                                                
# 92 "driver.mlg"
      merge_test ind1 
                                                                ) in fun ind1
                                                           ?loc ~atts ()
                                                           -> coqpp_body ind1
                                                           (Attributes.unsupported_attributes atts)), None))]

let () = Vernacextend.static_vernac_extend ~plugin:(Some "coq-quickchick.plugin") ~command:"Derive" ~classifier:(fun _ -> Vernacextend.classify_as_sideeff) ?entry:None 
         [(Vernacextend.TyML (false, Vernacextend.TyTerminal ("Derive", 
                                     Vernacextend.TyNonTerminal (Extend.TUentry (Genarg.get_arg_tag wit_constr), 
                                     Vernacextend.TyTerminal ("for", 
                                     Vernacextend.TyNonTerminal (Extend.TUentry (Genarg.get_arg_tag wit_constr), 
                                     Vernacextend.TyNil)))), (let coqpp_body class_name
                                                             inductive
                                                             () = Vernacextend.vtdefault (fun () -> 
                                                                  
# 97 "driver.mlg"
       dispatch class_name inductive "" "" 
                                                                  ) in fun class_name
                                                             inductive
                                                             ?loc ~atts ()
                                                             -> coqpp_body class_name
                                                             inductive
                                                             (Attributes.unsupported_attributes atts)), None));
         (Vernacextend.TyML (false, Vernacextend.TyTerminal ("Derive", 
                                    Vernacextend.TyNonTerminal (Extend.TUentry (Genarg.get_arg_tag wit_constr), 
                                    Vernacextend.TyTerminal ("for", Vernacextend.TyNonTerminal (
                                                                    Extend.TUentry (Genarg.get_arg_tag wit_constr), 
                                                                    Vernacextend.TyTerminal ("using", 
                                                                    Vernacextend.TyNonTerminal (
                                                                    Extend.TUentry (Genarg.get_arg_tag wit_ident), 
                                                                    Vernacextend.TyNil)))))), 
         (let coqpp_body class_name inductive genInst
         () = Vernacextend.vtdefault (fun () -> 
# 99 "driver.mlg"
       dispatch class_name inductive (Id.to_string genInst) ""
              ) in fun class_name
         inductive genInst ?loc ~atts () -> coqpp_body class_name inductive
         genInst (Attributes.unsupported_attributes atts)), None));
         (Vernacextend.TyML (false, Vernacextend.TyTerminal ("Derive", 
                                    Vernacextend.TyNonTerminal (Extend.TUentry (Genarg.get_arg_tag wit_constr), 
                                    Vernacextend.TyTerminal ("for", Vernacextend.TyNonTerminal (
                                                                    Extend.TUentry (Genarg.get_arg_tag wit_constr), 
                                                                    Vernacextend.TyTerminal ("using", 
                                                                    Vernacextend.TyNonTerminal (
                                                                    Extend.TUentry (Genarg.get_arg_tag wit_ident), 
                                                                    Vernacextend.TyTerminal ("and", 
                                                                    Vernacextend.TyNonTerminal (
                                                                    Extend.TUentry (Genarg.get_arg_tag wit_ident), 
                                                                    Vernacextend.TyNil)))))))), 
         (let coqpp_body class_name inductive genInst monInst
         () = Vernacextend.vtdefault (fun () -> 
# 101 "driver.mlg"
       dispatch class_name inductive (Id.to_string genInst) (Id.to_string monInst)
              ) in fun class_name
         inductive genInst monInst ?loc ~atts () -> coqpp_body class_name
         inductive genInst monInst (Attributes.unsupported_attributes atts)), None))]

let () = Vernacextend.static_vernac_extend ~plugin:(Some "coq-quickchick.plugin") ~command:"QCDerive" ~classifier:(fun _ -> Vernacextend.classify_as_sideeff) ?entry:None 
         [(Vernacextend.TyML (false, Vernacextend.TyTerminal ("QCDerive", 
                                     Vernacextend.TyNonTerminal (Extend.TUentry (Genarg.get_arg_tag wit_constr), 
                                     Vernacextend.TyTerminal ("for", 
                                     Vernacextend.TyNonTerminal (Extend.TUentry (Genarg.get_arg_tag wit_constr), 
                                     Vernacextend.TyNil)))), (let coqpp_body class_name
                                                             inductive
                                                             () = Vernacextend.vtdefault (fun () -> 
                                                                  
# 107 "driver.mlg"
       dispatch class_name inductive "" "" 
                                                                  ) in fun class_name
                                                             inductive
                                                             ?loc ~atts ()
                                                             -> coqpp_body class_name
                                                             inductive
                                                             (Attributes.unsupported_attributes atts)), None));
         (Vernacextend.TyML (false, Vernacextend.TyTerminal ("QCDerive", 
                                    Vernacextend.TyNonTerminal (Extend.TUentry (Genarg.get_arg_tag wit_constr), 
                                    Vernacextend.TyTerminal ("for", Vernacextend.TyNonTerminal (
                                                                    Extend.TUentry (Genarg.get_arg_tag wit_constr), 
                                                                    Vernacextend.TyTerminal ("using", 
                                                                    Vernacextend.TyNonTerminal (
                                                                    Extend.TUentry (Genarg.get_arg_tag wit_ident), 
                                                                    Vernacextend.TyNil)))))), 
         (let coqpp_body class_name inductive genInst
         () = Vernacextend.vtdefault (fun () -> 
# 109 "driver.mlg"
       dispatch class_name inductive (Id.to_string genInst) ""
              ) in fun class_name
         inductive genInst ?loc ~atts () -> coqpp_body class_name inductive
         genInst (Attributes.unsupported_attributes atts)), None));
         (Vernacextend.TyML (false, Vernacextend.TyTerminal ("QCDerive", 
                                    Vernacextend.TyNonTerminal (Extend.TUentry (Genarg.get_arg_tag wit_constr), 
                                    Vernacextend.TyTerminal ("for", Vernacextend.TyNonTerminal (
                                                                    Extend.TUentry (Genarg.get_arg_tag wit_constr), 
                                                                    Vernacextend.TyTerminal ("using", 
                                                                    Vernacextend.TyNonTerminal (
                                                                    Extend.TUentry (Genarg.get_arg_tag wit_ident), 
                                                                    Vernacextend.TyTerminal ("and", 
                                                                    Vernacextend.TyNonTerminal (
                                                                    Extend.TUentry (Genarg.get_arg_tag wit_ident), 
                                                                    Vernacextend.TyNil)))))))), 
         (let coqpp_body class_name inductive genInst monInst
         () = Vernacextend.vtdefault (fun () -> 
# 111 "driver.mlg"
       dispatch class_name inductive (Id.to_string genInst) (Id.to_string monInst)
              ) in fun class_name
         inductive genInst monInst ?loc ~atts () -> coqpp_body class_name
         inductive genInst monInst (Attributes.unsupported_attributes atts)), None))]

